use serde::Serialize;
use super::value::IOValueDomainCodec;
use super::value::boundary as B;
use super::value::writer::Writer;

pub use super::error::Error;
type Result<T> = std::result::Result<T, Error>;

#[derive(Debug)]
pub struct Serializer<'w, W: Writer> {
    pub write: &'w mut W,
}

impl<'w, W: Writer> Serializer<'w, W> {
    pub fn new(write: &'w mut W) -> Self {
        Serializer { write }
    }
}

enum SequenceVariant {
    Sequence,
    Record,
}

pub struct SerializeCompound<'a, 'w, W: Writer> {
    b: B::Type,
    i: B::Item,
    ser: &'a mut Serializer<'w, W>,
    variant: SequenceVariant,
}

pub struct SerializeDictionary<'a, 'w, W: Writer> {
    b: B::Type,
    ser: &'a mut Serializer<'w, W>,
}

impl<'a, 'w, W: Writer> serde::Serializer for &'a mut Serializer<'w, W> {
    type Ok = ();
    type Error = Error;
    type SerializeSeq = SerializeCompound<'a, 'w, W>;
    type SerializeTuple = SerializeCompound<'a, 'w, W>;
    type SerializeTupleStruct = SerializeCompound<'a, 'w, W>;
    type SerializeTupleVariant = SerializeCompound<'a, 'w, W>;
    type SerializeMap = SerializeDictionary<'a, 'w, W>;
    type SerializeStruct = SerializeCompound<'a, 'w, W>;
    type SerializeStructVariant = SerializeCompound<'a, 'w, W>;

    fn serialize_bool(self, v: bool) -> Result<Self::Ok> {
        Ok(self.write.write_bool(v)?)
    }

    fn serialize_i8(self, v: i8) -> Result<Self::Ok> {
        Ok(self.write.write_i8(v)?)
    }

    fn serialize_i16(self, v: i16) -> Result<Self::Ok> {
        Ok(self.write.write_i16(v)?)
    }

    fn serialize_i32(self, v: i32) -> Result<Self::Ok> {
        Ok(self.write.write_i32(v)?)
    }

    fn serialize_i64(self, v: i64) -> Result<Self::Ok> {
        Ok(self.write.write_i64(v)?)
    }

    fn serialize_u8(self, v: u8) -> Result<Self::Ok> {
        Ok(self.write.write_u8(v)?)
    }

    fn serialize_u16(self, v: u16) -> Result<Self::Ok> {
        Ok(self.write.write_u16(v)?)
    }

    fn serialize_u32(self, v: u32) -> Result<Self::Ok> {
        Ok(self.write.write_u32(v)?)
    }

    fn serialize_u64(self, v: u64) -> Result<Self::Ok> {
        Ok(self.write.write_u64(v)?)
    }

    fn serialize_f32(self, v: f32) -> Result<Self::Ok> {
        Ok(self.write.write_f32(v)?)
    }

    fn serialize_f64(self, v: f64) -> Result<Self::Ok> {
        Ok(self.write.write_f64(v)?)
    }

    fn serialize_char(self, v: char) -> Result<Self::Ok> {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol("UnicodeScalar")?;
        self.write.boundary(&B::mid(B::Item::RecordLabel, B::Item::RecordField))?;
        self.write.write_u32(v as u32)?;
        self.write.boundary(&B::end(B::Item::RecordField))?;
        Ok(self.write.end_record()?)
    }

    fn serialize_str(self, v: &str) -> Result<Self::Ok> {
        Ok(self.write.write_string(v)?)
    }

    fn serialize_bytes(self, v: &[u8]) -> Result<Self::Ok> {
        Ok(self.write.write_bytes(v)?)
    }

    fn serialize_none(self) -> Result<Self::Ok> {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol("None")?;
        self.write.boundary(&B::end(B::Item::RecordLabel))?;
        Ok(self.write.end_record()?)
    }

    fn serialize_some<T: ?Sized>(self, v: &T) -> Result<Self::Ok> where T: Serialize {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol("Some")?;
        self.write.boundary(&B::mid(B::Item::RecordLabel, B::Item::RecordField))?;
        to_writer(self.write, v)?;
        self.write.boundary(&B::end(B::Item::RecordField))?;
        Ok(self.write.end_record()?)
    }

    fn serialize_unit(self) -> Result<Self::Ok> {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol("tuple")?;
        self.write.boundary(&B::end(B::Item::RecordLabel))?;
        Ok(self.write.end_record()?)
    }

    fn serialize_unit_struct(self, name: &'static str) -> Result<Self::Ok> {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol(name)?;
        self.write.boundary(&B::end(B::Item::RecordLabel))?;
        Ok(self.write.end_record()?)
    }

    fn serialize_unit_variant(self,
                              _name: &'static str,
                              _variant: u32,
                              variant_name: &'static str) ->
        Result<Self::Ok>
    {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol(variant_name)?;
        self.write.boundary(&B::end(B::Item::RecordLabel))?;
        Ok(self.write.end_record()?)
    }

    fn serialize_newtype_struct<T: ?Sized>(self, name: &'static str, value: &T) ->
        Result<Self::Ok> where T: Serialize
    {
        match super::value::magic::receive_output_value(name, value) {
            Some(v) => Ok(self.write.write(&mut IOValueDomainCodec, &v)?),
            None => {
                // TODO: This is apparently discouraged, and we should apparently just serialize `value`?
                self.write.start_record()?;
                self.write.boundary(&B::start(B::Item::RecordLabel))?;
                self.write.write_symbol(name)?;
                self.write.boundary(&B::mid(B::Item::RecordLabel, B::Item::RecordField))?;
                to_writer(self.write, value)?;
                self.write.boundary(&B::end(B::Item::RecordField))?;
                Ok(self.write.end_record()?)
            }
        }
    }

    fn serialize_newtype_variant<T: ?Sized>(self,
                                            _name: &'static str,
                                            _variant: u32,
                                            variant_name: &'static str,
                                            value: &T) ->
        Result<Self::Ok> where T: Serialize
    {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol(variant_name)?;
        self.write.boundary(&B::mid(B::Item::RecordLabel, B::Item::RecordField))?;
        to_writer(self.write, value)?;
        self.write.boundary(&B::end(B::Item::RecordField))?;
        Ok(self.write.end_record()?)
    }

    fn serialize_seq(self, _count: Option<usize>) -> Result<Self::SerializeSeq> {
        self.write.start_sequence()?;
        Ok(SerializeCompound::seq(self))
    }

    fn serialize_tuple(self, _count: usize) -> Result<Self::SerializeTuple> {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol("tuple")?;
        Ok(SerializeCompound::rec(self))
    }

    fn serialize_tuple_struct(self, name: &'static str, _count: usize) ->
        Result<Self::SerializeTupleStruct>
    {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol(name)?;
        Ok(SerializeCompound::rec(self))
    }

    fn serialize_tuple_variant(self,
                               _name: &'static str,
                               _variant: u32,
                               variant_name: &'static str,
                               _count: usize) ->
        Result<Self::SerializeTupleVariant>
    {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol(variant_name)?;
        Ok(SerializeCompound::rec(self))
    }

    fn serialize_map(self, _count: Option<usize>) -> Result<Self::SerializeMap> {
        self.write.start_dictionary()?;
        Ok(SerializeDictionary { b: B::Type::default(), ser: self })
    }

    fn serialize_struct(self, name: &'static str, _count: usize) -> Result<Self::SerializeStruct> {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol(name)?;
        Ok(SerializeCompound::rec(self))
    }

    fn serialize_struct_variant(self,
                                _name: &'static str,
                                _variant: u32,
                                variant_name: &'static str,
                                _count: usize) ->
        Result<Self::SerializeStructVariant>
    {
        self.write.start_record()?;
        self.write.boundary(&B::start(B::Item::RecordLabel))?;
        self.write.write_symbol(variant_name)?;
        Ok(SerializeCompound::rec(self))
    }
}

impl<'a, 'w, W: Writer> serde::ser::SerializeMap for SerializeDictionary<'a, 'w, W> {
    type Ok = ();
    type Error = Error;

    fn serialize_key<T: ?Sized>(&mut self, key: &T) -> Result<()> where T: Serialize {
        self.b.opening = Some(B::Item::DictionaryKey);
        self.ser.write.boundary(&self.b)?;
        to_writer(self.ser.write, key)?;
        self.b.shift(None);
        Ok(())
    }

    fn serialize_value<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.b.opening = Some(B::Item::DictionaryValue);
        self.ser.write.boundary(&self.b)?;
        to_writer(self.ser.write, value)?;
        self.b.shift(None);
        Ok(())
    }

    fn end(self) -> Result<Self::Ok> {
        self.ser.write.boundary(&self.b)?;
        Ok(self.ser.write.end_dictionary()?)
    }
}

impl<'a, 'w, W: Writer> SerializeCompound<'a, 'w, W> {
    fn seq(ser: &'a mut Serializer<'w, W>) -> Self {
        SerializeCompound {
            b: B::Type::default(),
            i: B::Item::SequenceValue,
            ser,
            variant: SequenceVariant::Sequence,
        }
    }

    fn rec(ser: &'a mut Serializer<'w, W>) -> Self {
        SerializeCompound {
            b: B::end(B::Item::RecordLabel),
            i: B::Item::RecordField,
            ser,
            variant: SequenceVariant::Record,
        }
    }

    fn extend<T: ?Sized>(&mut self, value: &T) -> Result<()>
    where T: Serialize
    {
        self.b.opening = Some(self.i.clone());
        self.ser.write.boundary(&self.b)?;
        to_writer(self.ser.write, value)?;
        self.b.shift(None);
        Ok(())
    }

    fn complete(self) -> Result<()> {
        match self.variant {
            SequenceVariant::Sequence => {
                self.ser.write.boundary(&self.b)?;
                Ok(self.ser.write.end_sequence()?)
            }
            SequenceVariant::Record => {
                self.ser.write.boundary(&self.b)?;
                Ok(self.ser.write.end_record()?)
            }
        }
    }
}

impl<'a, 'w, W: Writer> serde::ser::SerializeStruct for SerializeCompound<'a, 'w, W> {
    type Ok = ();
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<()>
    where T: Serialize
    {
        self.extend(value)
    }

    fn end(self) -> Result<Self::Ok> {
        self.complete()
    }
}

impl<'a, 'w, W: Writer> serde::ser::SerializeStructVariant for SerializeCompound<'a, 'w, W> {
    type Ok = ();
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, _name: &'static str, value: &T) -> Result<()>
    where T: Serialize
    {
        self.extend(value)
    }

    fn end(self) -> Result<Self::Ok> {
        self.complete()
    }
}

impl<'a, 'w, W: Writer> serde::ser::SerializeTuple for SerializeCompound<'a, 'w, W> {
    type Ok = ();
    type Error = Error;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.extend(value)
    }

    fn end(self) -> Result<Self::Ok> {
        self.complete()
    }
}

impl<'a, 'w, W: Writer> serde::ser::SerializeTupleStruct for SerializeCompound<'a, 'w, W> {
    type Ok = ();
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.extend(value)
    }

    fn end(self) -> Result<Self::Ok> {
        self.complete()
    }
}

impl<'a, 'w, W: Writer> serde::ser::SerializeTupleVariant for SerializeCompound<'a, 'w, W> {
    type Ok = ();
    type Error = Error;

    fn serialize_field<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.extend(value)
    }

    fn end(self) -> Result<Self::Ok> {
        self.complete()
    }
}

impl<'a, 'w, W: Writer> serde::ser::SerializeSeq for SerializeCompound<'a, 'w, W> {
    type Ok = ();
    type Error = Error;

    fn serialize_element<T: ?Sized>(&mut self, value: &T) -> Result<()> where T: Serialize {
        self.extend(value)
    }

    fn end(self) -> Result<Self::Ok> {
        self.complete()
    }
}

pub fn to_writer<W: Writer, T: Serialize + ?Sized>(write: &mut W, value: &T) -> Result<()> {
    Ok(value.serialize(&mut Serializer::new(write))?)
}
